@@ -2,16 +2,14 @@
 """DB-SIG compliant module for communicating with MS SQL servers"""
 
 __author__ = 'Mikhail Denisenko <denisenkom@gmail.com>'
-__version__ = '0.4.0'
+__version__ = '0.5.0'
 
 import logging
 import decimal
-import datetime
 import re
 from tds import *
-from mem import *
+from tds import _TdsContext
 from login import *
-from config import *
 from query import *
 
 logger = logging.getLogger(__name__)
@@ -25,65 +23,6 @@ threadsafety = 1
 # this module uses extended python format codes
 paramstyle = 'pyformat'
 
-# stored procedure output parameter
-class output:
-    #property
-    def type(self):
-        """
-        This is the type of the parameter.
-        """
-        return self._type
-
-    @property
-    def value(self):
-        """
-        This is the value of the parameter.
-        """
-        return self._value
-
-
-    def __init__(self, param_type, value=None):
-        self._type = param_type
-        self._value = value
-
-#############################
-## DB-API type definitions ##
-#############################
-STRING = 1
-BINARY = 2
-NUMBER = 3
-DATETIME = 4
-DECIMAL = 5
-
-##################
-## DB-LIB types ##
-##################
-SQLBIT = SYBBIT
-SQLBITN = 104
-SQLCHAR = SYBCHAR
-SQLDATETIME = SYBDATETIME
-SQLDATETIM4 = SYBDATETIME4
-SQLDATETIMN = SYBDATETIMN
-SQLDECIMAL = SYBDECIMAL
-SQLFLT4 = SYBREAL
-SQLFLT8 = SYBFLT8
-SQLFLTN = SYBFLTN
-SQLIMAGE = SYBIMAGE
-SQLINT1 = SYBINT1
-SQLINT2 = SYBINT2
-SQLINT4 = SYBINT4
-SQLINT8 = SYBINT8
-SQLINTN = SYBINTN
-SQLMONEY = SYBMONEY
-SQLMONEY4 = SYBMONEY4
-SQLMONEYN = SYBMONEYN
-SQLNUMERIC = SYBNUMERIC
-SQLREAL = SYBREAL
-SQLTEXT = SYBTEXT
-SQLVARBINARY = SYBVARBINARY
-SQLVARCHAR = SYBVARCHAR
-SQLUUID = 36
-
 DB_RES_INIT            = 0
 DB_RES_RESULTSET_EMPTY = 1
 DB_RES_RESULTSET_ROWS  = 2
@@ -139,7 +78,6 @@ class MemoryChunkedHandler(object):
         return self.sio.getvalue()
 
 
-min_error_severity = 6
 ######################
 ## Connection class ##
 ######################
@@ -163,14 +101,6 @@ class Connection(object):
         """
         return self._autocommit
 
-    @property
-    def rows_affected(self):
-        """
-        Number of rows affected by last query. For SELECT statements this
-        value is only meaningful after reading all rows.
-        """
-        return self._rows_affected
-
     @property
     def chunk_handler(self):
         '''
@@ -188,7 +118,8 @@ class Connection(object):
         '''
         Returns version of tds protocol that is being used by this connection
         '''
-        return self.tds_socket.tds_version
+        tds = self._get_connection()
+        return tds.tds_version
 
     @property
     def product_version(self):
@@ -198,48 +129,29 @@ class Connection(object):
         return self.tds_socket.product_version
 
     def _get_connection(self):
+        if self._closed:
+            raise Error('Connection is closed')
         if self.tds_socket.is_dead():
             # clear transaction
             self.tds_socket.tds72_transaction = '\x00\x00\x00\x00\x00\x00\x00\x00'
             tds_connect_and_login(self.tds_socket, self._login)
-            # Set some connection properties to some reasonable values
-            # textsize - http://msdn.microsoft.com/en-us/library/aa259190%28v=sql.80%29.aspx
-            query = '''
-                SET ARITHABORT ON;
-                SET CONCAT_NULL_YIELDS_NULL ON;
-                SET ANSI_NULLS ON;
-                SET ANSI_NULL_DFLT_ON ON;
-                SET ANSI_PADDING ON;
-                SET ANSI_WARNINGS ON;
-                SET ANSI_NULL_DFLT_ON ON;
-                SET CURSOR_CLOSE_ON_COMMIT ON;
-                SET QUOTED_IDENTIFIER ON;
-                SET TEXTSIZE 2147483647;
-            '''
-            cur = self.cursor()
-            try:
-                cur.execute(query)
-            finally:
-                cur.close()
-            tds_submit_begin_tran(self.tds_socket)
+            self._try_activate_cursor(None)
+            if not self._autocommit:
+                tds_submit_begin_tran(self.tds_socket)
             self._sqlok()
         return self.tds_socket
 
     def __init__(self, server, user, password,
             charset, database, appname, port, tds_version,
-            as_dict, encryption_level, login_timeout, timeout):
-        self._autocommit = False
+            as_dict, encryption_level, login_timeout, timeout, autocommit):
+        self._autocommit = autocommit
         logger.debug("Connection.__init__()")
         self._charset = ''
-        self.last_msg_no = 0
-        self.last_msg_severity = 0
-        self.last_msg_state = 0
-        self.last_msg_str = ''
-        self.last_msg_srv = ''
-        self.last_msg_proc = ''
         self._as_dict = as_dict
         self._state = DB_RES_NO_MORE_RESULTS
         self.tds_socket = None
+        self._closed = False
+        self._active_cursor = None
 
         # support MS methods of connecting locally
         instance = ""
@@ -264,6 +176,13 @@ class Connection(object):
             login.tds_version = tds_version
         login.database = database
 
+        # that will set:
+        # ANSI_DEFAULTS to ON,
+        # IMPLICIT_TRANSACTIONS to OFF,
+        # TEXTSIZE to 0x7FFFFFFF (2GB) (TDS 7.2 and below), TEXTSIZE to infinite (introduced in TDS 7.3),
+        # and ROWCOUNT to infinite
+        login.option_flag2 = TDS_ODBC_ON
+
         # Set the character set name
         if charset:
             _charset = charset
@@ -273,14 +192,10 @@ class Connection(object):
         # Connect to the server
         login.server_name = server
         login.instance_name = instance
-        ctx = tds_alloc_context()
-        ctx.msg_handler = self._msg_handler
-        ctx.err_handler = self._err_handler
+        ctx = _TdsContext()
         ctx.int_handler = self._int_handler
         self.tds_socket = tds_alloc_socket(ctx, 512)
         self.tds_socket.chunk_handler = MemoryChunkedHandler()
-        login.option_flag2 &= ~0x02 # we're not an ODBC driver
-        tds_fix_login(login) # initialize from Environment variables
 
         login.connect_timeout = login_timeout
         login.query_timeout = timeout
@@ -309,10 +224,13 @@ class Connection(object):
         if self._autocommit:
             return
 
-        self.cancel()
+        tds = self._get_connection()
         try:
-            tds_submit_commit(self.tds_socket, True)
+            self._try_activate_cursor(None)
+            tds_submit_commit(tds, True)
             self._sqlok()
+            while self._nextset(None):
+                pass
         except Exception, e:
             raise OperationalError('Cannot commit transaction: ' + str(e[0]))
 
@@ -331,33 +249,17 @@ class Connection(object):
             return
 
         if not self.tds_socket.is_dead():
+            tds = self._get_connection()
             self.cancel()
-            tds_submit_rollback(self.tds_socket, True)
+            self._active_cursor = None
+            tds_submit_rollback(tds, True)
             self._sqlok()
-
-    def clr_err(self):
-        self.last_msg_no = 0
-        self.last_msg_severity = 0
-        self.last_msg_state = 0
+            while self._nextset(None):
+                pass
 
     def _int_handler(self):
         raise Exception('not implemented')
 
-    def _err_handler(self, tds_ctx, tds, msg):
-        self._msg_handler(tds_ctx, tds, msg)
-
-    def _msg_handler(self, tds_ctx, tds, msg):
-        if msg['severity'] < min_error_severity:
-            return
-        if msg['severity'] > self.last_msg_severity:
-            self.last_msg_severity = msg['severity']
-            self.last_msg_no = msg['msgno']
-            self.last_msg_state = msg['state']
-            self.last_msg_line = msg['line_number']
-            self.last_msg_str = msg['message']
-            self.last_msg_srv = msg['server']
-            self.last_msg_proc = msg['proc_name']
-
     def __del__(self):
         logger.debug("MSSQLConnection.__del__()")
         self.close()
@@ -371,8 +273,6 @@ class Connection(object):
         this case.
         """
         logger.debug("MSSQLConnection.cancel()")
-        self.clr_err()
-
         tds = self.tds_socket
         if not tds.is_dead():
             tds_send_cancel(tds)
@@ -387,7 +287,9 @@ class Connection(object):
         this case.
         """
         logger.debug("MSSQLConnection.close()")
-        self.clr_err()
+        if self._closed:
+            raise Error('Connection closed')
+        self._closed = True
         tds = self.tds_socket
         if tds is not None:
             tds_close_socket(tds)
@@ -424,7 +326,11 @@ class Connection(object):
         mask = TDS_STOPAT_ROWFMT|TDS_RETURN_DONE|TDS_RETURN_ROW|TDS_RETURN_COMPUTE
 
         # Get the row from the TDS stream.
-        rc, res_type, _ = tds_process_tokens(tds, mask)
+        rc, res_type, done_flags = tds_process_tokens(tds, mask)
+        if done_flags & TDS_DONE_ERROR:
+            raise_db_exception(tds)
+            assert False
+            raise Exception('FAIL')
         if rc == TDS_SUCCESS:
             if res_type in (TDS_ROW_RESULT, TDS_COMPUTE_RESULT):
                 # Add the row to the row buffer, whose capacity is always at least 1
@@ -442,18 +348,13 @@ class Connection(object):
         #CHECK_CONN(FAIL);
 
         tds = self.tds_socket
-        # See what the next packet from the server is.
-        # We want to skip any messages which are not processable. 
-        # We're looking for a result token or a done token.
         #
+        # If we hit an end token -- e.g. if the command
+        # submitted returned no data (like an insert) -- then
+        # we process the end token to extract the status code. 
+        #
+        logger.debug("dbsqlok() not done, calling tds_process_tokens()")
         while True:
-            #
-            # If we hit an end token -- e.g. if the command
-            # submitted returned no data (like an insert) -- then
-            # we process the end token to extract the status code. 
-            #
-            logger.debug("dbsqlok() not done, calling tds_process_tokens()")
-
             tds_code, result_type, done_flags = tds_process_tokens(tds, TDS_TOKEN_RESULTS)
 
             #
@@ -461,131 +362,48 @@ class Connection(object):
             # by a RAISERROR statement.  Microsoft db-lib returns FAIL in that case. 
             #/
             if done_flags & TDS_DONE_ERROR:
-                maybe_raise_MSSQLDatabaseException(self)
+                raise_db_exception(tds)
                 assert False
                 raise Exception('FAIL')
-            if tds_code == TDS_NO_MORE_RESULTS:
-                return
-
             if result_type == TDS_ROWFMT_RESULT:
-                pass
-            elif result_type == TDS_COMPUTEFMT_RESULT:
-                self._state = _DB_RES_RESULTSET_EMPTY;
-                logger.debug("dbsqlok() found result token")
-                break
-            elif result_type in (TDS_COMPUTE_RESULT, TDS_ROW_RESULT):
-                logger.debug("dbsqlok() found result token")
+                self._state = DB_RES_RESULTSET_ROWS
                 break
             elif result_type == TDS_DONEINPROC_RESULT:
-                pass
+                if not done_flags & TDS_DONE_COUNT:
+                    # skip results that don't event have rowcount
+                    continue
+                self._state = DB_RES_RESULTSET_EMPTY
+                break
             elif result_type in (TDS_DONE_RESULT, TDS_DONEPROC_RESULT):
-                logger.debug("dbsqlok() end status is %s", prdbresults_state(self._state))
-                if done_flags & TDS_DONE_ERROR:
-                    if done_flags & TDS_DONE_MORE_RESULTS:
-                        self._state = DB_RES_NEXT_RESULT
-                    else:
-                        self._state = DB_RES_NO_MORE_RESULTS
-
+                if done_flags & TDS_DONE_MORE_RESULTS:
+                    self._state = DB_RES_NEXT_RESULT
                 else:
-                    logger.debug("dbsqlok() end status was success")
-                    self._state = DB_RES_SUCCEED
+                    self._state = DB_RES_NO_MORE_RESULTS
+                break
+            elif result_type == TDS_STATUS_RESULT:
+                continue
             else:
-                logger.debug('logic error: tds_process_tokens result_type %d', result_type);
+                logger.error('logic error: tds_process_tokens result_type %d', result_type);
 
-    def _start_results(self):
-        result_type = 0
-
-        tds = self.tds_socket
-
-        logger.debug("_start_results begin: _state is %d (%s)\n", 
-                                        self._state, prdbresults_state(self._state))
-        if self._state == DB_RES_RESULTSET_ROWS:
-            raise Exception('FAIL')
-        elif self._state == DB_RES_NO_MORE_RESULTS:
-            return
-
-        while True:
-            retcode, result_type, done_flags = tds_process_tokens(tds, TDS_TOKEN_RESULTS)
-
-            logger.debug("_start_results() tds_process_tokens returned %d (%s),\n\t\t\tresult_type %s\n", 
-                                            retcode, prretcode(retcode), prresult_type(result_type))
-
-            if retcode == TDS_SUCCESS:
-                if result_type == TDS_ROWFMT_RESULT:
-                    self._state = DB_RES_RESULTSET_EMPTY
-
-                elif result_type == TDS_COMPUTEFMT_RESULT:
-                    pass
-
-                elif result_type in (TDS_ROW_RESULT, TDS_COMPUTE_RESULT):
-                    self._state = DB_RES_RESULTSET_ROWS
-                    return
-
-                elif result_type in (TDS_DONE_RESULT, TDS_DONEPROC_RESULT):
-                    logger.debug("_start_results() got done token _state is %d (%s)",
-                                    self._state, prdbresults_state(self._state))
-
-                    # A done token signifies the end of a logical command.
-                    # There are three possibilities:
-                    # 1. Simple command with no result set, i.e. update, delete, insert
-                    # 2. Command with result set but no rows
-                    # 3. Command with result set and rows
-                    #
-                    if self._state in (DB_RES_INIT, DB_RES_NEXT_RESULT):
-                        self._state = DB_RES_NEXT_RESULT
-                        if done_flags & TDS_DONE_ERROR:
-                            raise Exception('FAIL')
-
-                    elif self._state in (DB_RES_RESULTSET_EMPTY, DB_RES_RESULTSET_ROWS):
-                        self._state = DB_RES_NEXT_RESULT
-                        return
-                    else:
-                        assert False
-
-                elif result_type == TDS_DONEINPROC_RESULT:
-                        #
-                        # Return SUCCEED on a command within a stored procedure
-                        # only if the command returned a result set. 
-                        #
-                        if self._state in (DB_RES_INIT, DB_RES_NEXT_RESULT):
-                            self._state = DB_RES_NEXT_RESULT
-                        elif self._state in (DB_RES_RESULTSET_EMPTY, DB_RES_RESULTSET_ROWS):
-                            self._state = DB_RES_NEXT_RESULT
-                            return
-                        elif self._state in (DB_RES_NO_MORE_RESULTS, DB_RES_SUCCEED):
-                            pass
-
-            elif retcode == TDS_NO_MORE_RESULTS:
-                self._state = DB_RES_NO_MORE_RESULTS
-                return
-            else:
-                self._state = DB_RES_INIT
-                raise Exception('FAIL')
-
-    def _getrow(self, throw):
+    def _fetchone(self, cursor):
+        if self._active_cursor is not cursor:
+            raise Error('This cursor is not active')
         """
         Helper method used by fetchone and fetchmany to fetch and handle
         """
+        tds = self.tds_socket
+        if tds.res_info is None:
+            raise Error("Previous statement didn't produce any results")
+
         if self._state == DB_RES_NO_MORE_RESULTS:
-            logger.debug("MSSQLConnection.fetch_next_row(): NO MORE RESULTS")
-            if throw:
-                raise StopIteration
             return None
 
         self._nextrow()
 
-        check_cancel_and_raise(self)
-
         if self._state != DB_RES_RESULTSET_ROWS:
-            logger.debug("MSSQLConnection.fetch_next_row(): NO MORE ROWS")
-            # 'rows_affected' is nonzero only after all records are read
-            tds_socket = self.tds_socket
-            self._rows_affected = tds_socket.rows_affected
-            if throw:
-                raise StopIteration
             return None
 
-        cols = self.tds_socket.res_info.columns
+        cols = tds.res_info.columns
         row = tuple(col.value for col in cols)
         if self.as_dict:
             row_dict = dict(enumerate(cols))
@@ -593,6 +411,108 @@ class Connection(object):
             row = row_dict
         return row
 
+    def _nextset(self, cursor):
+        if cursor is not self._active_cursor:
+            raise Error('This cursor is not active')
+
+        while self._state == DB_RES_RESULTSET_ROWS:
+            self._nextrow()
+        self._sqlok()
+        return None if self._state == DB_RES_NO_MORE_RESULTS else True
+
+    def _rowcount(self, cursor):
+        if cursor is not self._active_cursor:
+            return -1
+        tds = self.tds_socket
+        return tds.rows_affected
+
+    def _get_proc_return_status(self, cursor):
+        if cursor is not self._active_cursor:
+            return None
+        tds = self.tds_socket
+        if not tds.has_status:
+            tds_process_tokens(tds, TDS_RETURN_PROC)
+        return tds.ret_status if tds.has_status else None
+
+    def _description(self, cursor):
+        if cursor is not self._active_cursor:
+            return None
+        res = self.tds_socket.res_info
+        if res:
+            return res.description
+        else:
+            return None
+
+    def _native_description(self):
+        if cursor is not self._active_cursor:
+            return None
+        res = self._conn.tds_socket.res_info
+        if res:
+            return res.native_descr
+        else:
+            return None
+
+    def _close_cursor(self, cursor):
+        if cursor is self._active_cursor:
+            self._active_cursor = None
+        cursor._conn = None
+
+    def _try_activate_cursor(self, cursor):
+        tds = self._get_connection()
+        if cursor is self._active_cursor or self._active_cursor is None:
+            self.cancel()
+        else:
+            if tds.state == TDS_PENDING:
+                rc, result_type, _ = tds_process_tokens(tds, TDS_TOKEN_TRAILING)
+                if rc != TDS_NO_MORE_RESULTS:
+                    raise InterfaceError('Results are still pending on connection')
+        self._active_cursor = cursor
+
+    def _execute(self, cursor, operation, params):
+        tds = self._get_connection()
+        self._try_activate_cursor(cursor)
+        tds_submit_query(tds, operation, params)
+        self._state = DB_RES_INIT
+        self._sqlok()
+
+    def _callproc(self, cursor, procname, parameters):
+        logger.debug('callproc begin')
+        tds = self._get_connection()
+        self._try_activate_cursor(cursor)
+        tds_submit_rpc(tds, procname, parameters)
+        tds.output_params = {}
+        self._state = DB_RES_INIT
+        while True:
+            tds_code, result_type, done_flags = tds_process_tokens(tds, TDS_TOKEN_RESULTS)
+            #
+            # The error flag may be set for any intervening DONEINPROC packet, in particular
+            # by a RAISERROR statement.  Microsoft db-lib returns FAIL in that case. 
+            #/
+            if done_flags & TDS_DONE_ERROR:
+                raise_db_exception(tds)
+                assert False
+                raise Exception('FAIL')
+            if result_type == TDS_ROWFMT_RESULT:
+                self._state = DB_RES_RESULTSET_ROWS
+                break
+            elif result_type == TDS_DONEINPROC_RESULT:
+                self._state = DB_RES_RESULTSET_EMPTY
+            elif result_type in (TDS_DONE_RESULT, TDS_DONEPROC_RESULT):
+                if done_flags & TDS_DONE_MORE_RESULTS:
+                    self._state = DB_RES_NEXT_RESULT
+                else:
+                    self._state = DB_RES_NO_MORE_RESULTS
+                break
+            elif result_type == TDS_STATUS_RESULT:
+                continue
+            else:
+                logger.error('logic error: tds_process_tokens result_type %d', result_type);
+        logger.debug('callproc end')
+        results = list(parameters)
+        for key, param in tds.output_params.items():
+            results[key] = param.value
+        return results
+
 ##################
 ## Cursor class ##
 ##################
@@ -601,22 +521,17 @@ class Cursor(object):
     This class represents a database cursor, which is used to issue queries
     and fetch results from a database connection.
     """
-    @property
-    def _source(self):
-        if self._conn == None:
-            raise InterfaceError('Cursor is closed.')
-        return self._conn
-
     def __init__(self, conn):
         self._conn = conn
         self._batchsize = 1
-        self._results = None
+        self.arraysize = 1
 
     def __enter__(self):
         return self
 
     def __exit__(self, *args):
-        self.close()
+        if self._conn is not None:
+            self.close()
 
     def __iter__(self):
         """
@@ -634,41 +549,22 @@ class Cursor(object):
         :keyword parameters: The optional parameters for the procedure
         :type parameters: sequence
         """
-        logger.debug('callproc begin')
-        self._results = None
-        tds = self._conn._get_connection()
-        if tds.state == TDS_PENDING:
-            rc, result_type, _ = tds_process_tokens(tds, TDS_TOKEN_TRAILING)
-            if rc != TDS_NO_MORE_RESULTS:
-                raise InterfaceError('Results are still pending on connection')
-        tds_submit_rpc(tds, procname, parameters)
-        self._source._state = DB_RES_INIT
-        self._source._sqlok()
-        check_cancel_and_raise(self._source)
-        logger.debug('callproc end')
+        return self._conn._callproc(self, procname, parameters)
+
+    @property
+    def return_value(self):
+        return self.get_proc_return_status()
 
     def get_proc_return_status(self):
-        self._get_results()
-        tds = self._conn.tds_socket
-        if not tds.has_status:
-            tds_process_tokens(tds, TDS_RETURN_PROC)
-        return tds.ret_status if tds.has_status else None
+        return self._conn._get_proc_return_status(self)
 
     def close(self):
         """
         Closes the cursor. The cursor is unusable from this point.
         """
-        self._conn.cancel()
-        self._conn = None
+        self._conn._close_cursor(self)
 
     def execute(self, operation, params=()):
-        self._results = None
-        tds = self._conn._get_connection()
-        if tds.state == TDS_PENDING:
-            rc, result_type, _ = tds_process_tokens(tds, TDS_TOKEN_TRAILING)
-            if rc != TDS_NO_MORE_RESULTS:
-                raise InterfaceError('Results are still pending on connection')
-
         # Execute the query
         if params:
             if isinstance(params, (list, tuple)):
@@ -685,16 +581,17 @@ class Cursor(object):
                 operation = operation % rename
             logger.debug('converted query: {0}'.format(operation))
             logger.debug('params: {0}'.format(params))
-        tds_submit_query(tds, operation, params)
-        self._conn._state = DB_RES_INIT
-        self._conn._sqlok()
-        check_cancel_and_raise(self._conn)
+        self._conn._execute(self, operation, params)
 
     def executemany(self, operation, params_seq):
-        self._results = None
+        counts = []
+        tds = self._conn.tds_socket
         for params in params_seq:
             self.execute(operation, params)
-            # support correct rowcount across multiple executes
+            if tds.rows_affected != -1:
+                counts.append(tds.rows_affected)
+        if counts:
+            tds.rows_affected = sum(counts)
 
     def execute_scalar(self, query_string, params=None):
         """
@@ -724,76 +621,42 @@ class Cursor(object):
         return row[0]
 
     def nextset(self):
-        self._results = None
-        self._conn.clr_err()
-
-        self._conn._nextrow()
-        check_cancel_and_raise(self._conn)
-
-        while self._conn._state == DB_RES_RESULTSET_ROWS:
-            self._conn._nextrow()
-            check_cancel_and_raise(self._conn)
-        self._get_results()
-        return None if self._conn._state == DB_RES_NO_MORE_RESULTS else True
-
-    def _get_results(self):
-        if not self._results:
-            tds = self._conn.tds_socket
-            self._conn._start_results()
-            check_cancel_and_raise(self._conn)
-            self._rows_affected = tds.rows_affected if tds.rows_affected != TDS_NO_COUNT else -1
-            if self._conn._state == DB_RES_NO_MORE_RESULTS:
-                descr = None
-            else:
-                header_tuple = []
-                for col in tds.res_info.columns:
-                    coltype = get_api_coltype(col.column_type)
-                    header_tuple.append((col.column_name, coltype, None, None, None, None, None))
-                descr = tuple(header_tuple)
-            logger.debug('_get_results set description: %s', descr)
-            self._results = {'description': descr}
-        return self._results
+        return self._conn._nextset(self)
 
     @property
     def rowcount(self):
-        self._get_results()
-        return self._rows_affected
+        return self._conn._rowcount(self)
 
     @property
     def description(self):
-        return self._get_results()['description']
+        return self._conn._description(self)
+
+    @property
+    def native_description(self):
+        return self._conn._native_description(self)
 
     def fetchone(self):
-        self._get_results()
-        return self._conn._getrow(throw=False)
+        return self._conn._fetchone(self)
 
     def fetchmany(self, size=None):
-        self._get_results()
         if size == None:
-            size = self._batchsize
-        self.batchsize = size
+            size = self.arraysize
 
         rows = []
         for i in xrange(size):
-            row = self._conn._getrow(throw=False)
+            row = self.fetchone()
             if not row:
                 break
             rows.append(row)
         return rows
 
     def fetchall(self):
-        self._get_results()
-        rows = []
-        while True:
-            row = self._conn._getrow(throw=False)
-            if not row:
-                break
-            rows.append(row)
-        return rows
+        return list(row for row in self)
 
     def next(self):
-        self._get_results()
-        row = self._conn._getrow(throw=True)
+        row = self.fetchone()
+        if row is None:
+            raise StopIteration
         return row
 
     def setinputsizes(self, sizes=None):
@@ -811,7 +674,7 @@ class Cursor(object):
 def connect(server='.', database='', user='', password='', timeout=0,
         login_timeout=60, charset=None, as_dict=False,
         host='', appname=None, port=None, tds_version=TDS74,
-        encryption_level=TDS_ENCRYPTION_OFF):
+        encryption_level=TDS_ENCRYPTION_OFF, autocommit=True):
     """
     Constructor for creating a connection to the database. Returns a
     Connection object.
@@ -855,75 +718,26 @@ def connect(server='.', database='', user='', password='', timeout=0,
 
     conn = Connection(server, user, password, charset, database,
         appname, port, tds_version=tds_version, as_dict=as_dict, login_timeout=login_timeout,
-        timeout=timeout, encryption_level=encryption_level)
+        timeout=timeout, encryption_level=encryption_level, autocommit=autocommit)
 
     return conn
 
-def get_last_msg_str(conn):
-    return conn.last_msg_str
+def Date(year, month, day):
+    return date(year, month, day)
 
-def get_last_msg_srv(conn):
-    return conn.last_msg_srv
+def DateFromTicks(ticks):
+    return date.fromtimestamp(ticks)
 
-def get_last_msg_proc(conn):
-    return conn.last_msg_proc
+def Time(hour, minute, second, microsecond=0):
+    from datetime import time
+    return time(hour, minute, second, microsecond)
 
-def get_last_msg_no(conn):
-    return conn.last_msg_no
+def TimeFromTicks(ticks):
+    import time
+    return Time(*time.localtime(ticks)[3:6])
 
-def get_last_msg_severity(conn):
-    return conn.last_msg_severity
+def Timestamp(year, month, day, hour, minute, second):
+    return datetime(year, month, day, hour, minute, second)
 
-def get_last_msg_state(conn):
-    return conn.last_msg_state
-
-def get_last_msg_line(conn):
-    return conn.last_msg_line
-
-def maybe_raise_MSSQLDatabaseException(conn):
-
-    if get_last_msg_severity(conn) < min_error_severity:
-        return 0
-
-    error_msg = get_last_msg_str(conn)
-    if len(error_msg) == 0:
-        error_msg = "Unknown error"
-
-    msg_no = get_last_msg_no(conn)
-    if msg_no in prog_errors:
-        ex = ProgrammingError(error_msg)
-    elif msg_no in integrity_errors:
-        ex = IntegrityError(error_msg)
-    else:
-        ex = OperationalError(error_msg)
-    ex.msg_no = msg_no
-    ex.text = error_msg
-    ex.srvname = get_last_msg_srv(conn)
-    ex.procname = get_last_msg_proc(conn)
-    ex.number = get_last_msg_no(conn)
-    ex.severity = get_last_msg_severity(conn)
-    ex.state = get_last_msg_state(conn)
-    ex.line = get_last_msg_line(conn)
-    conn.cancel()
-    conn.clr_err()
-    raise ex
-
-def check_cancel_and_raise(conn):
-    return maybe_raise_MSSQLDatabaseException(conn)
-
-######################
-## Helper Functions ##
-######################
-def get_api_coltype(coltype):
-    if coltype in (SQLBIT, SQLINT1, SQLINT2, SQLINT4, SQLINT8, SQLINTN,
-            SQLFLT4, SQLFLT8, SQLFLTN):
-        return NUMBER
-    elif coltype in (SQLMONEY, SQLMONEY4, SQLMONEYN, SQLNUMERIC,
-            SQLDECIMAL):
-        return DECIMAL
-    elif coltype in (SQLDATETIME, SQLDATETIM4, SQLDATETIMN):
-        return DATETIME
-    elif coltype in (SQLVARCHAR, SQLCHAR, SQLTEXT):
-        return STRING
-    else:
-        return BINARY
+def TimestampFromTicks(ticks):
+    return datetime.fromtimestamp(ticks)
