@@ -1,10 +1,7 @@
 import logging
 import traceback
-from read import *
 from tdsproto import *
-from iconv import *
-from mem import *
-from tds_checks import *
+from tds import _Column
 from data import *
 
 logger = logging.getLogger(__name__)
@@ -52,6 +49,7 @@ def tds_token_name(marker):
     return token_names.get(marker, '')
 
 def tds_process_default_tokens(tds, marker):
+    r = tds._reader
     logger.debug('tds_process_default_tokens() marker is {0:x}({1})'.format(marker, tds_token_name(marker)))
     if tds.is_dead():
         logger.debug('leaving tds_process_login_tokens() connection dead')
@@ -68,7 +66,7 @@ def tds_process_default_tokens(tds, marker):
         tds_process_msg(tds, marker)
     elif marker == TDS_CAPABILITY_TOKEN:
         # TODO split two part of capability and use it
-        tok_size = tds_get_smallint(tds)
+        tok_size = r.get_smallint()
         # vicm
         #
         # Sybase 11.0 servers return the wrong length in the capability packet, causing use to read
@@ -80,8 +78,8 @@ def tds_process_default_tokens(tds, marker):
             #pend = tds_conn(tds)->capabilities + TDS_MAX_CAPABILITY;
 
             #while True:
-            #    type = tds_get_byte(tds)
-            #    size = tds_get_byte(tds)
+            #    type = r.get_byte()
+            #    size = r.get_byte()
             #    if ((p + 2) > pend)
             #        break
             #    *p++ = type;
@@ -93,10 +91,10 @@ def tds_process_default_tokens(tds, marker):
             #    if type == 2:
             #        break
         else:
-            tds_conn(tds).capabilities = tds_get_n(tds, min(tok_size, TDS_MAX_CAPABILITY))
+            tds_conn(tds).capabilities = r.readall(min(tok_size, TDS_MAX_CAPABILITY))
             # PARAM_TOKEN can be returned inserting text in db, to return new timestamp
     elif marker == TDS_PARAM_TOKEN:
-        tds_unget_byte(tds)
+        r.unget_byte()
         return tds_process_param_result_tokens(tds)
     elif marker == TDS7_RESULT_TOKEN:
         return tds7_process_result(tds)
@@ -126,14 +124,14 @@ def tds_process_default_tokens(tds, marker):
         return tds_process_cursor_tokens(tds)
     elif marker in (TDS5_DYNAMIC_TOKEN, TDS_LOGINACK_TOKEN, TDS_ORDERBY_TOKEN, TDS_CONTROL_TOKEN):
         logger.debug("Eating %s token", tds_token_name(marker))
-        tds_skip_n(tds, tds_get_smallint(tds))
+        r.skip(r.get_smallint())
     elif marker == TDS_TABNAME_TOKEN: # used for FOR BROWSE query
         return tds_process_tabname(tds)
     elif marker == TDS_COLINFO_TOKEN:
         return tds_process_colinfo(tds, None, 0)
     elif marker == TDS_ORDERBY2_TOKEN:
         logger.debug("Eating %s token", tds_token_name(marker))
-        tds_skip_n(tds, tds_get_int(tds))
+        r.skip(r.get_int())
     elif marker == TDS_NBC_ROW_TOKEN:
         return tds_process_nbcrow(tds)
     else:
@@ -145,8 +143,6 @@ def tds_process_default_tokens(tds, marker):
 # tds_process_row() processes rows and places them in the row buffer.
 #
 def tds_process_row(tds):
-    CHECK_TDS_EXTRA(tds)
-
     info = tds.current_results
     if not info:
         raise Exception('TDS_FAIL')
@@ -162,6 +158,7 @@ def tds_process_row(tds):
 # NBC=null bitmap compression row
 # http://msdn.microsoft.com/en-us/library/dd304783(v=prot.20).aspx
 def tds_process_nbcrow(tds):
+    r = tds._reader
     info = tds.current_results
     if not info:
         raise Exception('TDS_FAIL')
@@ -170,7 +167,7 @@ def tds_process_nbcrow(tds):
 
     # reading bitarray for nulls, 1 represent null values for
     # corresponding fields
-    nbc = tds_get_n(tds, (len(info.columns) + 7) / 8)
+    nbc = r.readall((len(info.columns) + 7) / 8)
     for i, curcol in enumerate(info.columns):
         if ord(nbc[i/8]) & (1 << i%8):
             curcol.value = None
@@ -186,12 +183,13 @@ def tds_process_nbcrow(tds):
 #        Is NULL nothing is returned
 #
 def tds_process_end(tds, marker):
-    tmp = tds_get_smallint(tds)
-    state = tds_get_smallint(tds)
-    more_results = tmp & TDS_DONE_MORE_RESULTS != 0
-    was_cancelled = tmp & TDS_DONE_CANCELLED != 0
-    error = tmp & TDS_DONE_ERROR != 0
-    done_count_valid = tmp & TDS_DONE_COUNT != 0
+    r = tds._reader
+    status = r.get_usmallint()
+    cur_cmd = r.get_usmallint()
+    more_results = status & TDS_DONE_MORE_RESULTS != 0
+    was_cancelled = status & TDS_DONE_CANCELLED != 0
+    error = status & TDS_DONE_ERROR != 0
+    done_count_valid = status & TDS_DONE_COUNT != 0
     logger.debug('tds_process_end: more_results = {0}\n'
             '\t\twas_cancelled = {1}\n'
             '\t\terror = {2}\n'
@@ -200,7 +198,7 @@ def tds_process_end(tds, marker):
         tds.res_info.more_results = more_results
         if not tds.current_results:
             tds.current_results = tds.res_info
-    rows_affected = tds_get_int8(tds) if IS_TDS72_PLUS(tds) else tds_get_int(tds)
+    rows_affected = r.get_int8() if IS_TDS72_PLUS(tds) else r.get_int()
     logger.debug('\t\trows_affected = {0}'.format(rows_affected))
     if was_cancelled or (not more_results and not tds.in_cancel):
         logger.debug('tds_process_end() state set to TDS_IDLE')
@@ -211,37 +209,35 @@ def tds_process_end(tds, marker):
     if done_count_valid:
         tds.rows_affected = rows_affected
     else:
-        tds.rows_affected = None
-    return (TDS_CANCELLED if was_cancelled else TDS_SUCCESS), tmp
+        tds.rows_affected = -1
+    return (TDS_CANCELLED if was_cancelled else TDS_SUCCESS), status
 
 def tds_process_env_chg(tds):
-    size = tds_get_smallint(tds)
-    type = tds_get_byte(tds)
+    r = tds._reader
+    size = r.get_smallint()
+    type = r.get_byte()
     if type == TDS_ENV_SQLCOLLATION:
-        size = tds_get_byte(tds)
+        size = r.get_byte()
         logger.debug("tds_process_env_chg(): {0} bytes of collation data received".format(size))
-        logger.debug("tds.collation was {0}".format(tds.collation));
-        if size < 5:
-            tds.collation = tds_get_n(tds, size)
-        else:
-            tds.collation = tds_get_n(tds, 5)
-            tds_get_n(tds, size - 5)
-            lcid = (ord(tds.collation[0]) + (ord(tds.collation[1]) << 8) + (ord(tds.collation[2]) << 16)) & 0xfffff
-            tds7_srv_charset_changed(tds, ord(tds.collation[4]), lcid)
-        logger.debug("tds.collation now {0}".format(tds.collation));
+        logger.debug("tds.collation was {0}".format(tds.collation))
+        tds.collation = r.get_collation()
+        r.skip(size - 5)
+        #tds7_srv_charset_changed(tds, tds.collation)
+        logger.debug("tds.collation now {0}".format(tds.collation))
         # discard old one
-        tds_get_n(tds, tds_get_byte(tds))
+        r.skip(r.get_byte())
     elif type == TDS_ENV_BEGINTRANS:
-        size = tds_get_byte(tds)
-        tds.tds72_transaction = tds_get_n(tds, 8)
-        tds_get_n(tds, tds_get_byte(tds))
+        size = r.get_byte()
+        # TODO: parse transaction
+        tds.tds72_transaction = r.readall(8)
+        r.skip(r.get_byte())
     elif type == TDS_ENV_COMMITTRANS or type == TDS_ENV_ROLLBACKTRANS:
         tds.tds72_transaction = None
-        tds_get_n(tds, tds_get_byte(tds))
-        tds_get_n(tds, tds_get_byte(tds))
+        r.skip(r.get_byte())
+        r.skip(r.get_byte())
     elif type == TDS_ENV_PACKSIZE:
-        newval = tds_get_string(tds, tds_get_byte(tds))
-        oldval = tds_get_string(tds, tds_get_byte(tds))
+        newval = r.read_ucs2(r.get_byte())
+        oldval = r.read_ucs2(r.get_byte())
         new_block_size = int(newval)
         if new_block_size >= 512:
             logger.info("changing block size from {0} to {1}".format(oldval, new_block_size))
@@ -250,40 +246,42 @@ def tds_process_env_chg(tds):
             # size more than what we specified
             #
             # Reallocate buffer if possible (strange values from server or out of memory) use older buffer */
-            tds_realloc_socket(tds, new_block_size)
+            tds._writer.bufsize = new_block_size
     elif type == TDS_ENV_DATABASE:
-        newval = tds_get_string(tds, tds_get_byte(tds))
-        oldval = tds_get_string(tds, tds_get_byte(tds))
+        newval = r.read_ucs2(r.get_byte())
+        oldval = r.read_ucs2(r.get_byte())
         tds.env.database = newval
     elif type == TDS_ENV_LANG:
-        newval = tds_get_string(tds, tds_get_byte(tds))
-        oldval = tds_get_string(tds, tds_get_byte(tds))
+        newval = r.read_ucs2(r.get_byte())
+        oldval = r.read_ucs2(r.get_byte())
         tds.env.language = newval
     elif type == TDS_ENV_CHARSET:
-        newval = tds_get_string(tds, tds_get_byte(tds))
-        oldval = tds_get_string(tds, tds_get_byte(tds))
+        newval = r.read_ucs2(r.get_byte())
+        oldval = r.read_ucs2(r.get_byte())
         logger.debug("server indicated charset change to \"{0}\"\n".format(newval))
         tds.env.charset = newval
         tds_srv_charset_changed(tds, newval)
     elif type == TDS_ENV_DB_MIRRORING_PARTNER:
-        newval = tds_get_string(tds, tds_get_byte(tds))
-        oldval = tds_get_string(tds, tds_get_byte(tds))
+        newval = r.read_ucs2(r.get_byte())
+        oldval = r.read_ucs2(r.get_byte())
 
     else:
         # discard byte values, not still supported
         # TODO support them
         # discard new one
-        tds_get_n(tds, tds_get_byte(tds))
+        r.skip(r.get_byte())
         # discard old one
-        tds_get_n(tds, tds_get_byte(tds))
+        r.skip(r.get_byte())
 
 
 def tds_process_msg(tds, marker):
-    size = tds_get_smallint(tds)
+    r = tds._reader
+    size = r.get_smallint()
     msg = {}
-    msg['msgno'] = tds_get_int(tds)
-    msg['state'] = tds_get_byte(tds)
-    msg['severity'] = tds_get_byte(tds)
+    msg['marker'] = marker
+    msg['msgno'] = r.get_int()
+    msg['state'] = r.get_byte()
+    msg['severity'] = r.get_byte()
     msg['sql_state'] = None
     has_eed = False
     if marker == TDS_EED_TOKEN:
@@ -291,11 +289,11 @@ def tds_process_msg(tds, marker):
             msg['priv_msg_type'] = 0
         else:
             msg['priv_msg_type'] = 1
-        len_sqlstate = tds_get_byte(tds)
-        msg['sql_state'] = tds_get_n(tds, len_sqlstate)
-        has_eed = tds_get_byte(tds)
+        len_sqlstate = r.get_byte()
+        msg['sql_state'] = r.readall(len_sqlstate)
+        has_eed = r.get_byte()
         # junk status and transaction state
-        tds_get_smallint(tds)
+        r.get_smallint()
     elif marker == TDS_INFO_TOKEN:
         msg['priv_msg_type'] = 0
     elif marker == TDS_ERROR_TOKEN:
@@ -303,26 +301,26 @@ def tds_process_msg(tds, marker):
     else:
         logger.error('tds_process_msg() called with unknown marker "{0}"'.format(marker))
     logger.debug('tds_process_msg() reading message {0} from server'.format(msg['msgno']))
-    msg['message'] = tds_get_string(tds, tds_get_smallint(tds))
+    msg['message'] = r.read_ucs2(r.get_smallint())
     # server name
-    msg['server'] = tds_get_string(tds, tds_get_byte(tds))
+    msg['server'] = r.read_ucs2(r.get_byte())
     if not msg['server'] and tds.login:
         msg['server'] = tds.server_name
     # stored proc name if available
-    msg['proc_name'] = tds_get_string(tds, tds_get_byte(tds))
-    msg['line_number'] = tds_get_int(tds) if IS_TDS72_PLUS(tds) else tds_get_smallint(tds)
+    msg['proc_name'] = r.read_ucs2(r.get_byte())
+    msg['line_number'] = r.get_int() if IS_TDS72_PLUS(tds) else r.get_smallint()
     if not msg['sql_state']:
         #msg['sql_state'] = tds_alloc_lookup_sqlstate(tds, msg['msgno'])
         pass
     # in case extended error data is sent, we just try to discard it
     if has_eed:
         while True:
-            next_marker = tds_get_byte(tds)
+            next_marker = r.get_byte()
             if next_marker in (TDS5_PARAMFMT_TOKEN, TDS5_PARAMFMT2_TOKEN, TDS5_PARAMS_TOKEN):
                 tds_process_default_tokens(tds, next_marker)
             else:
                 break
-        tds_unget_byte(tds)
+        r.unget_byte()
 
     # call msg_handler
 
@@ -337,11 +335,7 @@ def tds_process_msg(tds, marker):
     else:
         # EED can be followed to PARAMFMT/PARAMS, do not store it in dynamic
         tds.cur_dyn = None
-        if tds_get_ctx(tds).msg_handler:
-            logger.debug('tds_process_msg() calling client msg handler')
-            tds_get_ctx(tds).msg_handler(tds_get_ctx(tds), tds, msg)
-        elif msg['msgno']:
-            logger.warn(u'Msg {msgno}, Severity {severity}, State {state}, Server {server}, Line {line_number}\n{message}'.format(**msg))
+    tds.messages.append(msg)
 
 _SERVER_TO_CLIENT_MAPPING = {
     0x07000000: TDS70,
@@ -354,17 +348,18 @@ _SERVER_TO_CLIENT_MAPPING = {
     }
 
 def tds_process_login_tokens(tds):
+    r = tds._reader
     succeed = False
     logger.debug('tds_process_login_tokens()')
     ver = {}
     while True:
-        marker = tds_get_byte(tds)
+        marker = r.get_byte()
         logger.debug('looking for login token, got  {0:x}({1})'.format(marker, tds_token_name(marker)))
         if marker == TDS_LOGINACK_TOKEN:
             tds.tds71rev1 = 0
-            size = tds_get_smallint(tds)
-            ack = tds_get_byte(tds)
-            version = tds_get_uint_be(tds)
+            size = r.get_smallint()
+            ack = r.get_byte()
+            version = r.get_uint_be()
             ver['reported'] = version
             tds.tds_version = _SERVER_TO_CLIENT_MAPPING[version]
             if tds.tds_version == TDS71rev1:
@@ -388,17 +383,19 @@ def tds_process_login_tokens(tds):
             logger.debug('server reports TDS version {0:x}'.format(version))
             # get server product name
             # ignore product name length, some servers seem to set it incorrectly
-            tds_get_byte(tds)
+            r.get_byte()
             product_version = 0
             size -= 10
             if IS_TDS7_PLUS(tds):
                 product_version = 0x80000000
-                tds.product_name = tds_get_string(tds, size/2)
+                tds.product_name = r.read_ucs2(size/2)
             elif IS_TDS5_PLUS(tds):
-                tds.product_name = tds_get_string(tds, size)
+                raise NotImplementedError()
+                #tds.product_name = tds_get_string(tds, size)
             else:
-                tds.product_name = tds_get_string(tds, size)
-            product_version = tds_get_uint_be(tds)
+                raise NotImplementedError()
+                #tds.product_name = tds_get_string(tds, size)
+            product_version = r.get_uint_be()
             # MSSQL 6.5 and 7.0 seem to return strange values for this
             # using TDS 4.2, something like 5F 06 32 FF for 6.50
             tds_conn(tds).product_version = product_version
@@ -472,27 +469,20 @@ def tds_process_login_tokens(tds):
 def tds_process_tokens(tds, flag):
     parent = {'result_type': 0, 'return_flag': 0}
     done_flags = 0
-    #TDSPARAMINFO *pinfo = NULL;
-    #TDSCOLUMN   *curcol;
-    #TDSRET rc;
     saved_rows_affected = tds.rows_affected
-    #TDS_INT ret_status;
     cancel_seen = 0
     import tds as tdsflags
+    r = tds._reader
 
     def SET_RETURN(ret, f):
         parent['result_type'] = ret
         parent['return_flag'] = getattr(tdsflags, 'TDS_RETURN_' + f) | getattr(tdsflags, 'TDS_STOPAT_' + f)
         if flag & getattr(tdsflags, 'TDS_STOPAT_' + f):
-            tds_unget_byte(tds)
+            r.unget_byte()
             logger.debug("tds_process_tokens::SET_RETURN stopping on current token")
             return False
         return True
 
-    CHECK_TDS_EXTRA(tds)
-
-    #tdsdump_log(TDS_DBG_FUNC, "tds_process_tokens(%p, %p, %p, 0x%x)\n", tds, result_type, done_flags, flag)
-
     if tds.state == TDS_IDLE:
         logger.debug("tds_process_tokens() state is COMPLETED")
         return TDS_NO_MORE_RESULTS, TDS_DONE_RESULT, done_flags
@@ -502,7 +492,7 @@ def tds_process_tokens(tds, flag):
     try:
         rc = TDS_SUCCESS
         while True:
-            marker = tds_get_byte(tds)
+            marker = r.get_byte()
             logger.info("processing result tokens.  marker is  {0:x}({1})".format(marker, tds_token_name(marker)))
             if marker == TDS7_RESULT_TOKEN:
                 #
@@ -512,18 +502,18 @@ def tds_process_tokens(tds, flag):
                 #
                 if tds.internal_sp_called == TDS_SP_CURSORFETCH:
                     rc = tds7_process_result(tds)
-                    marker = tds_get_byte(tds)
+                    marker = r.get_byte()
                     if marker != TDS_TABNAME_TOKEN:
-                        tds_unget_byte(tds);
+                        r.unget_byte();
                     else:
                         rc = tds_process_tabname(tds);
                 else:
                     if SET_RETURN(TDS_ROWFMT_RESULT, 'ROWFMT'):
                         rc = tds7_process_result(tds)
                         # handle browse information (if presents)
-                        marker = tds_get_byte(tds)
+                        marker = r.get_byte()
                         if marker != TDS_TABNAME_TOKEN:
-                            tds_unget_byte(tds)
+                            r.unget_byte()
                             rc = TDS_SUCCESS
                         else:
                             rc = tds_process_tabname(tds)
@@ -539,22 +529,22 @@ def tds_process_tokens(tds, flag):
                 if SET_RETURN(TDS_ROWFMT_RESULT, 'ROWFMT'):
                     rc = tds_process_col_fmt(tds)
                     # handle browse information (if present)
-                    marker = tds_get_byte(tds)
+                    marker = r.get_byte()
                     if marker == TDS_TABNAME_TOKEN:
                         rc = tds_process_tabname(tds)
                     else:
-                        tds_unget_byte(tds)
+                        r.unget_byte()
             elif marker == TDS_PARAM_TOKEN:
-                tds_unget_byte(tds)
+                r.unget_byte()
                 if tds.internal_sp_called:
                     logger.debug("processing parameters for sp {0}".formst(tds.internal_sp_called))
                     while True:
-                        marker = tds_get_byte(tds)
+                        marker = r.get_byte()
                         if marker != TDS_PARAM_TOKEN:
                             break
                         logger.debug("calling tds_process_param_result")
                         pinfo = tds_process_param_result(tds)
-                    tds_unget_byte(tds)
+                    r.unget_byte()
                     logger.debug("{0} hidden return parameters".format(pinfo.num_cols if pinfo else -1))
                     if pinfo and pinfo.num_cols > 0:
                         curcol = pinfo.columns[0]
@@ -605,8 +595,8 @@ def tds_process_tokens(tds, flag):
                 if SET_RETURN(TDS_COMPUTE_RESULT, 'COMPUTE'):
                     rc = tds_process_compute(tds, NULL)
             elif marker == TDS_RETURNSTATUS_TOKEN:
-                ret_status = tds_get_int(tds)
-                marker = tds_peek(tds)
+                ret_status = r.get_int()
+                marker = r.peek()
                 if marker in (TDS_PARAM_TOKEN, TDS_DONEPROC_TOKEN, TDS_DONE_TOKEN, TDS5_PARAMFMT_TOKEN, TDS5_PARAMFMT2_TOKEN):
                     if tds.internal_sp_called:
                         # TODO perhaps we should use ret_status ??
@@ -615,7 +605,7 @@ def tds_process_tokens(tds, flag):
                         # TODO optimize
                         flag &= ~TDS_STOPAT_PROC
                         if SET_RETURN(TDS_STATUS_RESULT, 'PROC'):
-                            tds.has_status = 1
+                            tds.has_status = True
                             tds.ret_status = ret_status
                             logger.debug("tds_process_tokens: return status is {0}".format(tds.ret_status))
                             rc = TDS_SUCCESS
@@ -624,20 +614,20 @@ def tds_process_tokens(tds, flag):
                 tds.cur_dyn = tds_process_dynamic(tds)
                 # special case, prepared statement cannot be prepared
                 if tds.cur_dyn and not tds.cur_dyn.emulated:
-                    marker = tds_get_byte(tds);
+                    marker = r.get_byte();
                     if marker == TDS_EED_TOKEN:
                         tds_process_msg(tds, marker)
                         if tds.cur_dyn and tds.cur_dyn.emulated:
-                            marker = tds_get_byte(tds)
+                            marker = r.get_byte()
                             if marker == TDS_DONE_TOKEN:
                                 rc, done_flags = tds_process_end(tds, marker)
                                 done_flags &= ~TDS_DONE_ERROR
                                 # FIXME warning to macro expansion
                                 SET_RETURN(TDS_DONE_RESULT, 'DONE')
                             else:
-                                tds_unget_byte(tds);
+                                r.unget_byte()
                     else:
-                        tds_unget_byte(tds);
+                        r.unget_byte()
             elif marker == TDS5_PARAMFMT_TOKEN:
                 if SET_RETURN(TDS_DESCRIBE_RESULT, 'PARAMFMT'):
                     rc = tds_process_dyn_result(tds)
@@ -717,8 +707,6 @@ def tds_process_tokens(tds, flag):
 # * At that point the connection should be ready to handle a new query.
 # */
 def tds_process_cancel(tds):
-    CHECK_TDS_EXTRA(tds);
-
     # silly cases, nothing to do
     if not tds.in_cancel:
         return TDS_SUCCESS
@@ -735,22 +723,32 @@ def tds_process_cancel(tds):
         elif rc in (TDS_CANCELLED, TDS_SUCCESS, TDS_NO_MORE_RESULTS):
             return TDS_SUCCESS
 
+def get_api_coltype(coltype):
+    if coltype in (SYBBIT, SYBINT1, SYBINT2, SYBINT4, SYBINT8, SYBINTN,
+            SYBREAL, SYBFLT8, SYBFLTN):
+        return NUMBER
+    elif coltype in (SYBMONEY, SYBMONEY4, SYBMONEYN, SYBNUMERIC,
+            SYBDECIMAL):
+        return DECIMAL
+    elif coltype in (SYBDATETIME, SYBDATETIME4, SYBDATETIMN):
+        return DATETIME
+    elif coltype in (SYBVARCHAR, SYBCHAR, SYBTEXT):
+        return STRING
+    else:
+        return BINARY
+
 #/**
 # * tds7_process_result() is the TDS 7.0 result set processing routine.  It 
 # * is responsible for populating the tds->res_info structure.
 # * This is a TDS 7.0 only function
 # */
 def tds7_process_result(tds):
-    #int col, num_cols;
-    #TDSRET result;
-    #TDSRESULTINFO *info;
-
-    CHECK_TDS_EXTRA(tds)
+    r = tds._reader
     logger.debug("processing TDS7 result metadata.")
 
     # read number of columns and allocate the columns structure
 
-    num_cols = tds_get_smallint(tds)
+    num_cols = r.get_smallint()
 
     # This can be a DUMMY results token from a cursor fetch
 
@@ -764,7 +762,6 @@ def tds7_process_result(tds):
     info = tds_alloc_results(num_cols)
     tds.current_results = info
     if tds.cur_cursor:
-        tds_free_results(tds.cur_cursor.res_info)
         tds.cur_cursor.res_info = info
         logger.debug("set current_results to cursor->res_info")
     else:
@@ -776,66 +773,59 @@ def tds7_process_result(tds):
     # server response
     #
     logger.debug("setting up {0} columns".format(num_cols))
+    header_tuple = []
     for col in range(num_cols):
         curcol = info.columns[col]
         tds7_get_data_info(tds, curcol)
-    if num_cols > 0:
-        dashes = "------------------------------"
-        #tdsdump_log(TDS_DBG_INFO1, " %-20s %-15s %-15s %-7s\n", "name", "size/wsize", "type/wtype", "utype");
-        #tdsdump_log(TDS_DBG_INFO1, " %-20s %15s %15s %7s\n", dashes+10, dashes+30-15, dashes+30-15, dashes+30-7);
-    for col in range(num_cols):
-        curcol = info.columns[col]
-
-        if curcol.column_name:
-            name = curcol.column_name
-        #tdsdump_log(TDS_DBG_INFO1, " %-20s %7d/%-7d %7d/%-7d %7d\n", 
-        #                                name, 
-        #                                curcol->column_size, curcol->on_server.column_size, 
-        #                                curcol->column_type, curcol->on_server.column_type, 
-        #                                curcol->column_usertype);
+        coltype = get_api_coltype(curcol.column_type)
+        precision = curcol.column_prec if hasattr(curcol, 'column_prec') else None
+        scale = curcol.column_scale if hasattr(curcol, 'column_scale') else None
+        header_tuple.append((curcol.column_name, coltype, None, None, precision, scale, curcol.column_nullable))
+    info.native_descr = tuple((col.column_name, col.column_type)
+            for col in tds.res_info.columns)
+    info.description = tuple(header_tuple)
 
     # all done now allocate a row for tds_process_row to use
     result = tds_alloc_row(info)
-    CHECK_TDS_EXTRA(tds)
     return result
 
+def tds_get_type_info(tds, curcol):
+    r = tds._reader
+    # User defined data type of the column
+    curcol.column_usertype = r.get_uint() if IS_TDS72_PLUS(tds) else r.get_usmallint()
+
+    curcol.column_flags = r.get_usmallint() # Flags
+
+    curcol.column_nullable = curcol.column_flags & 0x01;
+    curcol.column_writeable = (curcol.column_flags & 0x08) > 0
+    curcol.column_identity = (curcol.column_flags & 0x10) > 0
+
+    tds_set_column_type(tds, curcol, r.get_byte()) # sets "cardinal" type
+
+    curcol.column_timestamp = (curcol.column_type == SYBBINARY and curcol.column_usertype == TDS_UT_TIMESTAMP)
+
+    curcol.funcs.get_info(tds, curcol)
+
 #/**
 # * Read data information from wire
 # * \param tds state information for the socket and the TDS protocol
 # * \param curcol column where to store information
 # */
 def tds7_get_data_info(tds, curcol):
-    #int colnamelen;
-
-    CHECK_TDS_EXTRA(tds)
-    CHECK_COLUMN_EXTRA(curcol)
-
-    # User defined data type of the column
-    curcol.column_usertype = tds_get_int(tds) if IS_TDS72_PLUS(tds) else tds_get_smallint(tds)
-
-    curcol.column_flags = tds_get_smallint(tds) # Flags
-
-    curcol.column_nullable = curcol.column_flags & 0x01;
-    curcol.column_writeable = (curcol.column_flags & 0x08) > 0
-    curcol.column_identity = (curcol.column_flags & 0x10) > 0
-
-    tds_set_column_type(tds, curcol, tds_get_byte(tds)) # sets "cardinal" type
-
-    curcol.column_timestamp = (curcol.column_type == SYBBINARY and curcol.column_usertype == TDS_UT_TIMESTAMP)
-
-    curcol.funcs.get_info(tds, curcol)
+    r = tds._reader
+    tds_get_type_info(tds, curcol)
 
     # Adjust column size according to client's encoding
     #curcol.on_server.column_size = curcol.column_size
 
-    # NOTE adjustements must be done after curcol->char_conv initialization
+    # NOTE adjustements must be done after curcol.char_codec initialization
     adjust_character_column_size(tds, curcol)
 
     #
     # under 7.0 lengths are number of characters not
-    # number of bytes...tds_get_string handles this
+    # number of bytes... read_ucs2 handles this
     #
-    curcol.column_name = tds_get_string(tds, tds_get_byte(tds))
+    curcol.column_name = r.read_ucs2(r.get_byte())
 
     logger.debug("tds7_get_data_info: \n"
                 "\tcolname = %s (%d bytes)\n"
@@ -847,53 +837,31 @@ def tds7_get_data_info(tds, curcol):
                 curcol.on_server.column_type, tds_prtype(curcol.on_server.column_type),
                 curcol.column_varint_size))
 
-    CHECK_COLUMN_EXTRA(curcol)
-
     return TDS_SUCCESS
 
 #
 # Adjust column size according to client's encoding 
 #
 def adjust_character_column_size(tds, curcol):
-    CHECK_TDS_EXTRA(tds)
-    CHECK_COLUMN_EXTRA(curcol)
-
     if is_unicode_type(curcol.on_server.column_type):
-        curcol.char_conv = tds.char_convs[client2ucs2]
+        curcol.char_codec = ucs2_codec
 
     # Sybase UNI(VAR)CHAR fields are transmitted via SYBLONGBINARY and in UTF-16
     if curcol.on_server.column_type == SYBLONGBINARY and \
             curcol.column_usertype in (USER_UNICHAR_TYPE, USER_UNIVARCHAR_TYPE):
-        sybase_utf = "UTF-16LE"
-
-        curcol.char_conv = tds_iconv_get(tds, tds.char_convs[client2ucs2].client_charset.name, sybase_utf)
-
-        # fallback to UCS-2LE
-        # FIXME should be useless. Does not works always
-        if not curcol.char_conv:
-            curcol.char_conv = tds.char_convs[client2ucs2]
 
+        curcol.char_codec = ucs2_codec
     # FIXME: and sybase ??
-    if not curcol.char_conv and IS_TDS7_PLUS(tds) and is_ascii_type(curcol.on_server.column_type):
-        curcol.char_conv = tds.char_convs[client2server_chardata]
+    if not curcol.char_codec and IS_TDS7_PLUS(tds) and is_ascii_type(curcol.on_server.column_type):
+        curcol.char_codec = tds.collation.get_codec()
 
-    if not USE_ICONV(tds) or not curcol.char_conv:
+    if not curcol.char_codec:
         return
 
     curcol.on_server.column_size = curcol.column_size;
-    curcol.column_size = determine_adjusted_size(curcol.char_conv, curcol.column_size)
+    curcol.column_size = determine_adjusted_size(curcol.char_codec, curcol.column_size)
 
-    logger.debug("adjust_character_column_size:\n"
-                 "\tServer charset: %s\n"
-                 "\tServer column_size: %d\n"
-                 "\tClient charset: %s\n"
-                 "\tClient column_size: %d\n", 
-                 curcol.char_conv['server_charset']['name'], 
-                 curcol.on_server.column_size, 
-                 curcol.char_conv['client_charset']['name'],
-                 curcol.column_size)
-
-def determine_adjusted_size(char_conv, column_size):
+def determine_adjusted_size(char_codec, column_size):
     return column_size
 
 _prtype_map = dict((
@@ -946,10 +914,19 @@ _prtype_map = dict((
 def tds_prtype(token):
     return _prtype_map.get(token, '')
 
-def tds_swap_numeric(num):
-    # swap the sign
-    arr_sign = chr(0) if ord(num.array[0]) else chr(1)
-    # swap the data
-    arr_prec = tds_swap_bytes(num.array[1:], tds_numeric_bytes_per_prec[num.precision] - 1)
-    arr_rest = num.array[1+tds_numeric_bytes_per_prec[num.precision] - 1:]
-    return ''.join([arr_sign, arr_prec, arr_rest])
+def tds_process_param_result_tokens(tds):
+    r = tds._reader
+    while True:
+        token = r.get_byte()
+        if token == TDS_PARAM_TOKEN:
+            ordinal = r.get_usmallint()
+            name = r.read_ucs2(r.get_byte())
+            status = r.get_byte() # 1 - OUTPUT of sp, 2 - result of udf
+            param = _Column()
+            param.column_name = name
+            tds_get_type_info(tds, param)
+            param.funcs.get_data(tds, param)
+            tds.output_params[ordinal] = param
+        else:
+            r.unget_byte()
+            return
